﻿using EonaCat.HID.Helpers;
using System;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.HID
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/License for full license details.
    public class Device : IDevice
    {
        public event InsertedEventHandler OnInserted;

        public event RemovedEventHandler OnRemoved;

        private const int BUFFER_SIZE = 254;
        private readonly string _description;
        private readonly string _path;
        private readonly Attributes _attributes;

        private readonly Capabilities _capabilities;
        private readonly DeviceEventMonitor _deviceEventMonitor;

        private bool _monitorDeviceEvents;

        protected delegate DeviceData ReadDelegate(int timeout);

        protected delegate Report ReadReportDelegate(int timeout);

        private delegate bool WriteDelegate(byte[] data, int timeout);

        private delegate bool WriteReportDelegate(Report report, int timeout);

        internal Device(string devicePath, string description = null)
        {
            _deviceEventMonitor = new DeviceEventMonitor(this);
            _deviceEventMonitor.Inserted += DeviceEventMonitorInserted;
            _deviceEventMonitor.Removed += DeviceEventMonitorRemoved;

            _path = devicePath;
            _description = description;

            try
            {
                _deviceModeHelper = new DeviceModeHelper();

                var handle = OpenDeviceIO(_path, NativeMethods.ACCESS_NONE);
                _attributes = GetDeviceAttributes(handle);
                _capabilities = GetDeviceCapabilities(handle);
                CloseDeviceIO(handle);
            }
            catch (Exception exception)
            {
                throw new Exception($"EonaCat HID: Error querying device '{devicePath}'.", exception);
            }
        }

        public bool IsOpen { get; private set; }
        public bool IsConnected
        { get { return Devices.IsConnected(_path); } }
        public string Description
        { get { return _description; } }
        public Capabilities Capabilities
        { get { return _capabilities; } }
        public Attributes Info
        { get { return _attributes; } }
        public string DevicePath
        { get { return _path; } }

        public bool MonitorDeviceEvents
        {
            get { return _monitorDeviceEvents; }
            set
            {
                if (value & !_monitorDeviceEvents)
                {
                    _deviceEventMonitor.Init();
                }

                _monitorDeviceEvents = value;
            }
        }

        private DeviceModeHelper _deviceModeHelper { get; }

        public override string ToString()
        {
            return $"VendorID={_attributes.VendorIdHex}, ProductID={_attributes.ProductIdHex}, Version={_attributes.Version}, DevicePath={_path}";
        }

        public void OpenDevice()
        {
            OpenDevice(DeviceMode.NonOverlapped, DeviceMode.NonOverlapped, ShareMode.ShareRead | ShareMode.ShareWrite);
        }

        public void OpenDevice(DeviceMode readMode, DeviceMode writeMode, ShareMode shareMode)
        {
            if (IsOpen)
            {
                return;
            }

            _deviceModeHelper.ReadMode = readMode;
            _deviceModeHelper.WriteMode = writeMode;
            _deviceModeHelper.ShareMode = shareMode;

            try
            {
                _deviceModeHelper.ReadHandle = OpenDeviceIO(_path, readMode, NativeMethods.GENERIC_READ, shareMode);
                _deviceModeHelper.WriteHandle = OpenDeviceIO(_path, writeMode, NativeMethods.GENERIC_WRITE, shareMode);
            }
            catch (Exception exception)
            {
                IsOpen = false;
                throw new Exception("EonaCat HID: Error opening device.", exception);
            }

            IsOpen = (_deviceModeHelper.ReadHandle.ToInt32() != NativeMethods.INVALID_HANDLE_VALUE && _deviceModeHelper.WriteHandle.ToInt32() != NativeMethods.INVALID_HANDLE_VALUE);
        }

        public void CloseDevice()
        {
            if (!IsOpen)
            {
                return;
            }

            CloseDeviceIO(_deviceModeHelper.ReadHandle);
            CloseDeviceIO(_deviceModeHelper.WriteHandle);
            IsOpen = false;
        }

        public DeviceData Read()
        {
            return Read(0);
        }

        public DeviceData Read(int timeout)
        {
            if (IsConnected)
            {
                if (!IsOpen)
                {
                    OpenDevice(_deviceModeHelper.ReadMode, _deviceModeHelper.WriteMode, _deviceModeHelper.ShareMode);
                }

                try
                {
                    return ReadData(timeout);
                }
                catch
                {
                    return new DeviceData(DeviceData.ReadStatus.ReadError);
                }
            }
            return new DeviceData(DeviceData.ReadStatus.NotConnected);
        }

        public void Read(ReadCallback callback)
        {
            Read(callback, 0);
        }

        public void Read(ReadCallback callback, int timeout)
        {
            var readDelegate = new ReadDelegate(Read);
            var asyncState = new StateAsync(readDelegate, callback);
            readDelegate.BeginInvoke(timeout, EndRead, asyncState);
        }

        public async Task<DeviceData> ReadAsync(int timeout = 0)
        {
            var readDelegate = new ReadDelegate(Read);
#if NET20 || NET35 || NET5_0_OR_GREATER
            return await Task<DeviceData>.Factory.StartNew(() => readDelegate.Invoke(timeout));
#else
            return await Task<DeviceData>.Factory.FromAsync(readDelegate.BeginInvoke, readDelegate.EndInvoke, timeout, null);
#endif
        }

        public Report ReadReport(int timeout = 0)
        {
            return new Report(Capabilities.InputReportByteLength, Read(timeout));
        }

        public void ReadReport(ReadReportCallback callback, int timeout = 0)
        {
            var readReportDelegate = new ReadReportDelegate(ReadReport);
            var asyncState = new StateAsync(readReportDelegate, callback);
            readReportDelegate.BeginInvoke(timeout, EndReadReport, asyncState);
        }

        public async Task<Report> ReadReportAsync(int timeout = 0)
        {
            var readReportDelegate = new ReadReportDelegate(ReadReport);
#if NET20 || NET35 || NET5_0_OR_GREATER
            return await Task<Report>.Factory.StartNew(() => readReportDelegate.Invoke(timeout));
#else
            return await Task<Report>.Factory.FromAsync(readReportDelegate.BeginInvoke, readReportDelegate.EndInvoke, timeout, null);
#endif
        }

        public Report ReadReportSync(byte reportId)
        {
            byte[] commandBuffer = new byte[Capabilities.InputReportByteLength];
            commandBuffer[0] = reportId;
            bool bSuccess = NativeMethods.HidD_GetInputReport(_deviceModeHelper.ReadHandle, commandBuffer, commandBuffer.Length);
            DeviceData deviceData = new DeviceData(commandBuffer, bSuccess ? DeviceData.ReadStatus.Success : DeviceData.ReadStatus.NoDataRead);
            return new Report(Capabilities.InputReportByteLength, deviceData);
        }

        public bool ReadFeatureData(out byte[] data, byte reportId = 0)
        {
            if (_capabilities.FeatureReportByteLength <= 0)
            {
                data = new byte[0];
                return false;
            }

            data = new byte[_capabilities.FeatureReportByteLength];

            var buffer = CreateFeatureOutputBuffer();
            buffer[0] = reportId;

            IntPtr hidHandle = IntPtr.Zero;
            bool success = false;
            try
            {
                if (IsOpen)
                {
                    hidHandle = _deviceModeHelper.ReadHandle;
                }
                else
                {
                    hidHandle = OpenDeviceIO(_path, NativeMethods.ACCESS_NONE);
                }

                success = NativeMethods.HidD_GetFeature(hidHandle, buffer, buffer.Length);

                if (success)
                {
                    Array.Copy(buffer, 0, data, 0, Math.Min(data.Length, _capabilities.FeatureReportByteLength));
                }
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error accessing device '{0}'.", _path), exception);
            }
            finally
            {
                if (hidHandle != IntPtr.Zero && hidHandle != _deviceModeHelper.ReadHandle)
                {
                    CloseDeviceIO(hidHandle);
                }
            }

            return success;
        }

        public string ReadProductName()
        {
            if (ReadProductName(out byte[] bytes))
            {
                return ByteHelper.GetFilledBytesString(bytes);
            }
            return string.Empty;
        }

        public bool ReadProductName(out byte[] data)
        {
            data = new byte[BUFFER_SIZE];
            IntPtr handle = IntPtr.Zero;
            bool success = false;
            try
            {
                if (IsOpen)
                {
                    handle = _deviceModeHelper.ReadHandle;
                }
                else
                {
                    handle = OpenDeviceIO(_path, NativeMethods.ACCESS_NONE);
                }

                success = NativeMethods.HidD_GetProductString(handle, data, data.Length);
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error accessing device '{0}'.", _path), exception);
            }
            finally
            {
                if (handle != IntPtr.Zero && handle != _deviceModeHelper.ReadHandle)
                {
                    CloseDeviceIO(handle);
                }
            }

            return success;
        }

        public string ReadManufacturer()
        {
            if (ReadManufacturer(out byte[] bytes))
            {
                return ByteHelper.GetFilledBytesString(bytes);
            }
            return string.Empty;
        }

        public bool ReadManufacturer(out byte[] data)
        {
            data = new byte[BUFFER_SIZE];
            IntPtr handle = IntPtr.Zero;
            bool success = false;
            try
            {
                if (IsOpen)
                {
                    handle = _deviceModeHelper.ReadHandle;
                }
                else
                {
                    handle = OpenDeviceIO(_path, NativeMethods.ACCESS_NONE);
                }

                success = NativeMethods.HidD_GetManufacturerString(handle, data, data.Length);
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error accessing device '{0}'.", _path), exception);
            }
            finally
            {
                if (handle != IntPtr.Zero && handle != _deviceModeHelper.ReadHandle)
                {
                    CloseDeviceIO(handle);
                }
            }

            return success;
        }

        public string ReadSerialNumber()
        {
            if (ReadSerialNumber(out byte[] bytes))
            {
                return ByteHelper.GetFilledBytesString(bytes);
            }
            return string.Empty;
        }

        public bool ReadSerialNumber(out byte[] data)
        {
            data = new byte[BUFFER_SIZE];
            IntPtr hidHandle = IntPtr.Zero;
            bool success = false;
            try
            {
                if (IsOpen)
                {
                    hidHandle = _deviceModeHelper.ReadHandle;
                }
                else
                {
                    hidHandle = OpenDeviceIO(_path, NativeMethods.ACCESS_NONE);
                }

                success = NativeMethods.HidD_GetSerialNumberString(hidHandle, data, data.Length);
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error accessing device '{0}'.", _path), exception);
            }
            finally
            {
                if (hidHandle != IntPtr.Zero && hidHandle != _deviceModeHelper.ReadHandle)
                {
                    CloseDeviceIO(hidHandle);
                }
            }

            return success;
        }

        public bool Write(byte[] data)
        {
            return Write(data, 0);
        }

        public bool Write(byte[] data, int timeout)
        {
            if (IsConnected)
            {
                if (!IsOpen)
                {
                    OpenDevice(_deviceModeHelper.ReadMode, _deviceModeHelper.WriteMode, _deviceModeHelper.ShareMode);
                }

                try
                {
                    return WriteData(data, timeout);
                }
                catch
                {
                    return false;
                }
            }
            return false;
        }

        public void Write(byte[] data, WriteCallback callback)
        {
            Write(data, callback, 0);
        }

        public void Write(byte[] data, WriteCallback callback, int timeout)
        {
            var writeDelegate = new WriteDelegate(Write);
            var asyncState = new StateAsync(writeDelegate, callback);
            writeDelegate.BeginInvoke(data, timeout, EndWrite, asyncState);
        }

        public async Task<bool> WriteAsync(byte[] data, int timeout = 0)
        {
            var writeDelegate = new WriteDelegate(Write);
#if NET20 || NET35 || NET5_0_OR_GREATER
            return await Task<bool>.Factory.StartNew(() => writeDelegate.Invoke(data, timeout));
#else
            return await Task<bool>.Factory.FromAsync(writeDelegate.BeginInvoke, writeDelegate.EndInvoke, data, timeout, null);
#endif
        }

        public bool WriteReport(Report report)
        {
            return WriteReport(report, 0);
        }

        public bool WriteReport(Report report, int timeout)
        {
            return Write(report.GetBytes(), timeout);
        }

        public void WriteReport(Report report, WriteCallback callback)
        {
            WriteReport(report, callback, 0);
        }

        public void WriteReport(Report report, WriteCallback callback, int timeout)
        {
            var writeReportDelegate = new WriteReportDelegate(WriteReport);
            var asyncState = new StateAsync(writeReportDelegate, callback);
            writeReportDelegate.BeginInvoke(report, timeout, EndWriteReport, asyncState);
        }

        /// <summary>
        /// Handle data transfers on the control channel.
        /// This method places data on the control channel for devices
        /// that do not support the interupt transfers
        /// </summary>
        /// <param name="report">The outbound HID report</param>
        /// <returns>The result of the tranfer request: true if successful otherwise false</returns>
        ///
        public bool WriteReportSync(Report report)
        {
            if (null != report)
            {
                byte[] buffer = report.GetBytes();
                return (NativeMethods.HidD_SetOutputReport(_deviceModeHelper.WriteHandle, buffer, buffer.Length));
            }
            else
            {
                throw new ArgumentException("The output report is null, it must be allocated before you call this method", "report");
            }
        }

        public async Task<bool> WriteReportAsync(Report report, int timeout = 0)
        {
            var writeReportDelegate = new WriteReportDelegate(WriteReport);
#if NET20 || NET35 || NET5_0_OR_GREATER
            return await Task<bool>.Factory.StartNew(() => writeReportDelegate.Invoke(report, timeout));
#else
            return await Task<bool>.Factory.FromAsync(writeReportDelegate.BeginInvoke, writeReportDelegate.EndInvoke, report, timeout, null);
#endif
        }

        public Report CreateReport()
        {
            return new Report(Capabilities.OutputReportByteLength);
        }

        public bool WriteFeatureData(byte[] data)
        {
            if (_capabilities.FeatureReportByteLength <= 0)
            {
                return false;
            }

            var buffer = CreateFeatureOutputBuffer();

            Array.Copy(data, 0, buffer, 0, Math.Min(data.Length, _capabilities.FeatureReportByteLength));

            IntPtr hidHandle = IntPtr.Zero;
            bool success = false;
            try
            {
                if (IsOpen)
                {
                    hidHandle = _deviceModeHelper.WriteHandle;
                }
                else
                {
                    hidHandle = OpenDeviceIO(_path, NativeMethods.ACCESS_NONE);
                }

                //var overlapped = new NativeOverlapped();
                success = NativeMethods.HidD_SetFeature(hidHandle, buffer, buffer.Length);
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error accessing device '{0}'.", _path), exception);
            }
            finally
            {
                if (hidHandle != IntPtr.Zero && hidHandle != _deviceModeHelper.WriteHandle)
                {
                    CloseDeviceIO(hidHandle);
                }
            }
            return success;
        }

        protected static void EndRead(IAsyncResult ar)
        {
            var hidAsyncState = (StateAsync)ar.AsyncState;
            var callerDelegate = (ReadDelegate)hidAsyncState.CallerDelegate;
            var callbackDelegate = (ReadCallback)hidAsyncState.CallbackDelegate;
            var data = callerDelegate.EndInvoke(ar);

            if ((callbackDelegate != null))
            {
                callbackDelegate.Invoke(data);
            }
        }

        protected static void EndReadReport(IAsyncResult ar)
        {
            var hidAsyncState = (StateAsync)ar.AsyncState;
            var callerDelegate = (ReadReportDelegate)hidAsyncState.CallerDelegate;
            var callbackDelegate = (ReadReportCallback)hidAsyncState.CallbackDelegate;
            var report = callerDelegate.EndInvoke(ar);

            if ((callbackDelegate != null))
            {
                callbackDelegate.Invoke(report);
            }
        }

        private static void EndWrite(IAsyncResult ar)
        {
            var hidAsyncState = (StateAsync)ar.AsyncState;
            var callerDelegate = (WriteDelegate)hidAsyncState.CallerDelegate;
            var callbackDelegate = (WriteCallback)hidAsyncState.CallbackDelegate;
            var result = callerDelegate.EndInvoke(ar);

            if ((callbackDelegate != null))
            {
                callbackDelegate.Invoke(result);
            }
        }

        private static void EndWriteReport(IAsyncResult ar)
        {
            var hidAsyncState = (StateAsync)ar.AsyncState;
            var callerDelegate = (WriteReportDelegate)hidAsyncState.CallerDelegate;
            var callbackDelegate = (WriteCallback)hidAsyncState.CallbackDelegate;
            var result = callerDelegate.EndInvoke(ar);

            if ((callbackDelegate != null))
            {
                callbackDelegate.Invoke(result);
            }
        }

        private byte[] CreateInputBuffer()
        {
            return CreateBuffer(Capabilities.InputReportByteLength - 1);
        }

        private byte[] CreateOutputBuffer()
        {
            return CreateBuffer(Capabilities.OutputReportByteLength - 1);
        }

        private byte[] CreateFeatureOutputBuffer()
        {
            return CreateBuffer(Capabilities.FeatureReportByteLength - 1);
        }

        private static byte[] CreateBuffer(int length)
        {
            byte[] buffer = null;
            Array.Resize(ref buffer, length + 1);
            return buffer;
        }

        private static Attributes GetDeviceAttributes(IntPtr hidHandle)
        {
            var deviceAttributes = default(NativeMethods.HIDD_ATTRIBUTES);
            deviceAttributes.Size = Marshal.SizeOf(deviceAttributes);
            NativeMethods.HidD_GetAttributes(hidHandle, ref deviceAttributes);
            return new Attributes(deviceAttributes);
        }

        private static Capabilities GetDeviceCapabilities(IntPtr hidHandle)
        {
            var capabilities = default(NativeMethods.Capabilities);
            var preparsedDataPointer = default(IntPtr);

            if (NativeMethods.HidD_GetPreparsedData(hidHandle, ref preparsedDataPointer))
            {
                NativeMethods.HidP_GetCaps(preparsedDataPointer, ref capabilities);
                NativeMethods.HidD_FreePreparsedData(preparsedDataPointer);
            }
            return new Capabilities(capabilities);
        }

        private bool WriteData(byte[] data, int timeout)
        {
            if (_capabilities.OutputReportByteLength <= 0)
            {
                return false;
            }

            var buffer = CreateOutputBuffer();
            uint bytesWritten;

            Array.Copy(data, 0, buffer, 0, Math.Min(data.Length, _capabilities.OutputReportByteLength));

            if (_deviceModeHelper.WriteMode == DeviceMode.Overlapped)
            {
                var security = new NativeMethods.SECURITY_ATTRIBUTES();
                var overlapped = new NativeOverlapped();

                var overlapTimeout = timeout <= 0 ? NativeMethods.WAIT_INFINITE : timeout;

                security.lpSecurityDescriptor = IntPtr.Zero;
                security.bInheritHandle = true;
                security.nLength = Marshal.SizeOf(security);

                overlapped.OffsetLow = 0;
                overlapped.OffsetHigh = 0;
                overlapped.EventHandle = NativeMethods.CreateEvent(ref security, Convert.ToInt32(false), Convert.ToInt32(true), "");

                try
                {
                    NativeMethods.WriteFile(_deviceModeHelper.WriteHandle, buffer, (uint)buffer.Length, out bytesWritten, ref overlapped);
                }
                catch { return false; }

                var result = NativeMethods.WaitForSingleObject(overlapped.EventHandle, overlapTimeout);

                switch (result)
                {
                    case NativeMethods.WAIT_OBJECT_0:
                        return true;

                    case NativeMethods.WAIT_TIMEOUT:
                        return false;

                    case NativeMethods.WAIT_FAILED:
                        return false;

                    default:
                        return false;
                }
            }
            else
            {
                try
                {
                    var overlapped = new NativeOverlapped();
                    return NativeMethods.WriteFile(_deviceModeHelper.WriteHandle, buffer, (uint)buffer.Length, out bytesWritten, ref overlapped);
                }
                catch { return false; }
            }
        }

        protected DeviceData ReadData(int timeout)
        {
            var buffer = new byte[] { };
            var status = DeviceData.ReadStatus.NoDataRead;
            IntPtr nonManagedBuffer;

            if (_capabilities.InputReportByteLength > 0)
            {
                uint bytesRead;

                buffer = CreateInputBuffer();
                nonManagedBuffer = Marshal.AllocHGlobal(buffer.Length);

                if (_deviceModeHelper.ReadMode == DeviceMode.Overlapped)
                {
                    var security = new NativeMethods.SECURITY_ATTRIBUTES();
                    var overlapped = new NativeOverlapped();
                    var overlapTimeout = timeout <= 0 ? NativeMethods.WAIT_INFINITE : timeout;

                    security.lpSecurityDescriptor = IntPtr.Zero;
                    security.bInheritHandle = true;
                    security.nLength = Marshal.SizeOf(security);

                    overlapped.OffsetLow = 0;
                    overlapped.OffsetHigh = 0;
                    overlapped.EventHandle = NativeMethods.CreateEvent(ref security, Convert.ToInt32(false), Convert.ToInt32(true), string.Empty);

                    try
                    {
                        var success = NativeMethods.ReadFile(_deviceModeHelper.ReadHandle, nonManagedBuffer, (uint)buffer.Length, out bytesRead, ref overlapped);

                        if (success)
                        {
                            status = DeviceData.ReadStatus.Success; // No check here to see if bytesRead > 0 . Perhaps not necessary?
                        }
                        else
                        {
                            var result = NativeMethods.WaitForSingleObject(overlapped.EventHandle, overlapTimeout);

                            switch (result)
                            {
                                case NativeMethods.WAIT_OBJECT_0:
                                    status = DeviceData.ReadStatus.Success;
                                    NativeMethods.GetOverlappedResult(_deviceModeHelper.ReadHandle, ref overlapped, out bytesRead, false);
                                    break;

                                case NativeMethods.WAIT_TIMEOUT:
                                    status = DeviceData.ReadStatus.WaitTimedOut;
                                    buffer = new byte[] { };
                                    break;

                                case NativeMethods.WAIT_FAILED:
                                    status = DeviceData.ReadStatus.WaitFail;
                                    buffer = new byte[] { };
                                    break;

                                default:
                                    status = DeviceData.ReadStatus.NoDataRead;
                                    buffer = new byte[] { };
                                    break;
                            }
                        }
                        Marshal.Copy(nonManagedBuffer, buffer, 0, (int)bytesRead);
                    }
                    catch { status = DeviceData.ReadStatus.ReadError; }
                    finally
                    {
                        CloseDeviceIO(overlapped.EventHandle);
                        Marshal.FreeHGlobal(nonManagedBuffer);
                    }
                }
                else
                {
                    try
                    {
                        var overlapped = new NativeOverlapped();

                        NativeMethods.ReadFile(_deviceModeHelper.ReadHandle, nonManagedBuffer, (uint)buffer.Length, out bytesRead, ref overlapped);
                        status = DeviceData.ReadStatus.Success;
                        Marshal.Copy(nonManagedBuffer, buffer, 0, (int)bytesRead);
                    }
                    catch { status = DeviceData.ReadStatus.ReadError; }
                    finally { Marshal.FreeHGlobal(nonManagedBuffer); }
                }
            }
            return new DeviceData(buffer, status);
        }

        private static IntPtr OpenDeviceIO(string devicePath, uint deviceAccess)
        {
            return OpenDeviceIO(devicePath, DeviceMode.NonOverlapped, deviceAccess, ShareMode.ShareRead | ShareMode.ShareWrite);
        }

        private static IntPtr OpenDeviceIO(string devicePath, DeviceMode deviceMode, uint deviceAccess, ShareMode shareMode)
        {
            var security = new NativeMethods.SECURITY_ATTRIBUTES();
            var flags = 0;

            if (deviceMode == DeviceMode.Overlapped)
            {
                flags = NativeMethods.FILE_FLAG_OVERLAPPED;
            }

            security.lpSecurityDescriptor = IntPtr.Zero;
            security.bInheritHandle = true;
            security.nLength = Marshal.SizeOf(security);

            return NativeMethods.CreateFile(devicePath, deviceAccess, (int)shareMode, ref security, NativeMethods.OPEN_EXISTING, flags, hTemplateFile: IntPtr.Zero);
        }

        private static void CloseDeviceIO(IntPtr handle)
        {
            if (Environment.OSVersion.Version.Major > 5)
            {
                NativeMethods.CancelIoEx(handle, IntPtr.Zero);
            }
            NativeMethods.CloseHandle(handle);
        }

        private void DeviceEventMonitorInserted()
        {
            if (!IsOpen)
            {
                OpenDevice(_deviceModeHelper.ReadMode, _deviceModeHelper.WriteMode, _deviceModeHelper.ShareMode);
            }

            OnInserted?.Invoke();
        }

        private void DeviceEventMonitorRemoved()
        {
            if (IsOpen)
            {
                CloseDevice();
            }

            OnRemoved?.Invoke();
        }

        public void Dispose()
        {
            if (MonitorDeviceEvents)
            {
                MonitorDeviceEvents = false;
            }

            if (IsOpen)
            {
                CloseDevice();
            }
        }
    }
}