﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;

namespace EonaCat.HID
{
    public class Devices
    {
        private static Guid _hidClassGuid = Guid.Empty;

        public static bool IsConnected(string devicePath)
        {
            return EnumerateDevices().Any(x => x.Path == devicePath);
        }

        public static Device GetDevice(string devicePath)
        {
            return Enumerate(devicePath).FirstOrDefault();
        }

        public static IEnumerable<Device> Enumerate()
        {
            return EnumerateDevices().Select(x => new Device(x.Path, x.Description));
        }

        public static IEnumerable<Device> Enumerate(string devicePath)
        {
            return EnumerateDevices().Where(x => x.Path == devicePath).Select(x => new Device(x.Path, x.Description));
        }

        public static IEnumerable<Device> Enumerate(int vendorId, params int[] productIds)
        {
            return EnumerateDevices().Select(x => new Device(x.Path, x.Description)).Where(x => x.Info.VendorId == vendorId &&
                                                                                  productIds.Contains(x.Info.ProductId));
        }

        public static IEnumerable<Device> Enumerate(int vendorId, int productId, ushort UsagePage)
        {
            return EnumerateDevices().Select(x => new Device(x.Path, x.Description)).Where(x => x.Info.VendorId == vendorId &&
                                                                                  productId == (ushort)x.Info.ProductId && (ushort)x.Capabilities.UsagePage == UsagePage);
        }

        public static IEnumerable<Device> Enumerate(int vendorId)
        {
            return EnumerateDevices().Select(x => new Device(x.Path, x.Description)).Where(x => x.Info.VendorId == vendorId);
        }

        public static IEnumerable<Device> Enumerate(ushort UsagePage)
        {
            return EnumerateDevices().Select(x => new Device(x.Path, x.Description)).Where(x => (ushort)x.Capabilities.UsagePage == UsagePage);
        }

        internal class DeviceInfo
        { public string Path { get; set; } public string Description { get; set; } }

        internal static IEnumerable<DeviceInfo> EnumerateDevices()
        {
            var devices = new List<DeviceInfo>();
            var hidClass = HidClassGuid;
            var deviceInfoSet = NativeMethods.SetupDiGetClassDevs(ref hidClass, null, hwndParent: IntPtr.Zero, NativeMethods.DIGCF_PRESENT | NativeMethods.DIGCF_DEVICEINTERFACE);

            if (deviceInfoSet.ToInt64() != NativeMethods.INVALID_HANDLE_VALUE)
            {
                var deviceInfoData = CreateDeviceInfoData();
                var deviceIndex = 0;

                while (NativeMethods.SetupDiEnumDeviceInfo(deviceInfoSet, deviceIndex, ref deviceInfoData))
                {
                    deviceIndex += 1;

                    var deviceInterfaceData = new NativeMethods.SP_DEVICE_INTERFACE_DATA();
                    deviceInterfaceData.cbSize = Marshal.SizeOf(deviceInterfaceData);
                    var deviceInterfaceIndex = 0;

                    while (NativeMethods.SetupDiEnumDeviceInterfaces(deviceInfoSet, ref deviceInfoData, ref hidClass, deviceInterfaceIndex, ref deviceInterfaceData))
                    {
                        deviceInterfaceIndex++;
                        var devicePath = GetDevicePath(deviceInfoSet, deviceInterfaceData);
                        var description = GetBusReportedDeviceDescription(deviceInfoSet, ref deviceInfoData) ??
                                          GetDeviceDescription(deviceInfoSet, ref deviceInfoData);
                        devices.Add(new DeviceInfo { Path = devicePath, Description = description });
                    }
                }
                NativeMethods.SetupDiDestroyDeviceInfoList(deviceInfoSet);
            }
            return devices;
        }

        private static NativeMethods.SP_DEVINFO_DATA CreateDeviceInfoData()
        {
            var deviceInfoData = new NativeMethods.SP_DEVINFO_DATA();

            deviceInfoData.cbSize = Marshal.SizeOf(deviceInfoData);
            deviceInfoData.DevInst = 0;
            deviceInfoData.ClassGuid = Guid.Empty;
            deviceInfoData.Reserved = IntPtr.Zero;

            return deviceInfoData;
        }

        private static string GetDevicePath(IntPtr deviceInfoSet, NativeMethods.SP_DEVICE_INTERFACE_DATA deviceInterfaceData)
        {
            var bufferSize = 0;
            var interfaceDetail = new NativeMethods.SP_DEVICE_INTERFACE_DETAIL_DATA { Size = IntPtr.Size == 4 ? 4 + Marshal.SystemDefaultCharSize : 8 };

            NativeMethods.SetupDiGetDeviceInterfaceDetail(deviceInfoSet, ref deviceInterfaceData, IntPtr.Zero, 0, ref bufferSize, IntPtr.Zero);

            return NativeMethods.SetupDiGetDeviceInterfaceDetail(deviceInfoSet, ref deviceInterfaceData, ref interfaceDetail, bufferSize, ref bufferSize, IntPtr.Zero) ?
                interfaceDetail.DevicePath : null;
        }

        private static Guid HidClassGuid
        {
            get
            {
                if (_hidClassGuid.Equals(Guid.Empty))
                {
                    NativeMethods.HidD_GetHidGuid(ref _hidClassGuid);
                }

                return _hidClassGuid;
            }
        }

        private static string GetDeviceDescription(IntPtr deviceInfoSet, ref NativeMethods.SP_DEVINFO_DATA devinfoData)
        {
            unsafe
            {
                const int charCount = 1024;
                var descriptionBuffer = stackalloc char[charCount];

                var requiredSize = 0;
                var type = 0;

                if (NativeMethods.SetupDiGetDeviceRegistryProperty(deviceInfoSet,
                    ref devinfoData,
                    NativeMethods.SPDRP_DEVICEDESC,
                    ref type,
                    descriptionBuffer,
                    propertyBufferSize: charCount * sizeof(char),
                    ref requiredSize))
                {
                    return new string(descriptionBuffer);
                }

                return null;
            }
        }

        private static string GetBusReportedDeviceDescription(IntPtr deviceInfoSet, ref NativeMethods.SP_DEVINFO_DATA devinfoData)
        {
            unsafe
            {
                const int charCount = 1024;
                var descriptionBuffer = stackalloc char[charCount];

                if (Environment.OSVersion.Version.Major > 5)
                {
                    uint propertyType = 0;
                    var requiredSize = 0;

                    if (NativeMethods.SetupDiGetDeviceProperty(deviceInfoSet,
                        ref devinfoData,
                        ref NativeMethods.DEVPKEY_Device_BusReportedDeviceDesc,
                        ref propertyType,
                        descriptionBuffer,
                        propertyBufferSize: charCount * sizeof(char),
                        ref requiredSize,
                        0))
                    {
                        return new string(descriptionBuffer);
                    }
                }

                return null;
            }
        }
    }
}